

UMLS-Interface Configuration Information
  Sources:
    SNOMEDCT
  Relations:
    CHD
    PAR
  Database: 
    umls (MMSYS-2009AA-20090325)




The related (rb) CUIs to hand (C0018563):
  entire upper limb (body structure) (C1269078)
The related (rb) CUIs to hand (C1281583):
  entire upper limb (body structure) (C1269078)
